/*
 * xwxinput.c - thread to listen for input on X11 and inject into win32
 * Copyright (C) 2001 Kim Saunders, Sanjay Rao
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <wingdi.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <pthread.h>
#include "xwx.h"
#include "xwxmemory.h"
#include "xwxlines.h"
#include "xwxmouse.h"
#include "xwxkeyboard.h"
#include "xwxinput.h"

static pthread_t pthread;

static void expose(struct xinput *xinput, XEvent *evt) {
	/* TODO, replace the win32 Event's with pthread equivs.
	 * note that there is probably some bug involved with using
	 * pthread_mutex's, so it'd have to be a pthread_cond or so 
	 *
	 * alternately, replace the pthread_mutex's with CriticalSections and 
	 * just use win32
	 *
	 * for now, it's an odd mix of pthreads and win32
	 *
	 * see xwxlines.c for an example of how the other side should lock
	 * for exposures. The problem is how we queue up multiple exposure
	 * events. For now, there is no formal queue, so it passes one 
	 * along just fine, but waits if there's already one that's unprocessed.
	 * */

pthread_mutex_lock(&xinput->inputlock);
if (xinput->exp_b) {
	ResetEvent(xinput->deadlock);
	pthread_mutex_unlock(&xinput->inputlock);

	WaitForSingleObject(xinput->deadlock,INFINITE);

	pthread_mutex_lock(&xinput->inputlock);
}

#define XEXPOSE(a) ((XExposeEvent*)(a))
	xinput->exp_x=XEXPOSE(evt)->x;
	xinput->exp_y=XEXPOSE(evt)->y;
	xinput->exp_w=XEXPOSE(evt)->width;
	xinput->exp_h=XEXPOSE(evt)->height;

	xinput->exp_b=1;
#undef XEXPOSE
pthread_mutex_unlock(&xinput->inputlock);

}

static void setunfocus(struct xinput *xinput, int mode) {
pthread_mutex_lock(&xinput->inputlock);
xinput->unfocus=mode;
pthread_mutex_unlock(&xinput->inputlock);
}

static void *listen_input(void *xinput_in) {
struct xinput *xinput;
XEvent evt;
int sync=0;

xinput=(struct xinput*)xinput_in;

while (1) {
	/* process events from the X server */

	pthread_mutex_lock(&xinput->inputlock);
	if (!XCheckWindowEvent(xinput->x11info->display,
			xinput->x11info->window,
			ExposureMask |
			ButtonPressMask |
			ButtonReleaseMask |
			KeyPressMask |
			KeyReleaseMask |
			FocusChangeMask |
			PointerMotionMask,
			&evt)) {

		pthread_mutex_unlock(&xinput->inputlock);
		Sleep(25);
		continue;
	}
	pthread_mutex_unlock(&xinput->inputlock);

	{

		switch (evt.type) {
		/* if this is a focus event, sync on next key event */
			case FocusIn:
				setunfocus(xinput,0);
				sync=1;
				break;
			case FocusOut:
				setunfocus(xinput,1);
				sync=1;
				break;
			case KeyPress:
			case KeyRelease:
				if (sync) {
					keySyncLocks(xinput,xinput->x11info, 
						xinput->params, evt);
					sync=0;
				}
				keyPressEvent(xinput,xinput->x11info, 
						xinput->params, 
						evt);
				break;
			case Expose:
				expose(xinput,&evt);
				break;
			case ButtonPress:
			case ButtonRelease:
				mouUpdate(xinput,xinput->x11info);
				mouPressEvent(xinput->x11info, xinput->params, 
						evt);
				break;
			case MotionNotify:
				mouUpdate(xinput,xinput->x11info); 
				break;
		}
	}
}

return NULL;
}

int init_input(struct params *params, struct x11info *x11info,
		struct xinput *xinput) {

pthread_mutex_init(&xinput->inputlock,NULL);
xinput->deadlock=CreateEvent(NULL,TRUE,FALSE,NULL);
if (!xinput->deadlock) goto error;

xinput->x11info=x11info;
xinput->params=params;

if (pthread_create(&pthread,NULL,listen_input, (void *)xinput)) goto error;

return 0;
error:
	fprintf(stderr,"Error initializing input\n");
	return -1;
}

int deinit_input(struct xinput *xinput) {
pthread_cancel(&pthread);
pthread_mutex_unlock(&xinput->inputlock);
pthread_mutex_destroy(&xinput->inputlock);
CloseHandle(xinput->deadlock);
return 0;
}
